(* Global types, constants and subroutines for
   ExtFileTools, AesLib, FtpFileTools

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1.0 - Jun. 2006
   Vers. 2.0 - Apr. 2007 : Error messages changed
   last changed : March 2014
   *)

unit FtGlobal;

interface

type
  TFileAction = (acNone,acCopy,acCompress,acUnCompress,acEncrypt,acDecrypt,
                 acVerify,acAddVerify,acFtpWrite,acFtpRead,acScan);
  TProgressEvent = procedure(AAction : TFileAction; ACount: int64; ASpeedLimit : boolean) of object;
  TTotalProgressEvent = procedure(AAction : TFileAction; ACount: int64) of object;
  TTimeOutEvent = function : boolean of object;

const
  // error codes
  errOK         = 0;
  errFileCreate = 1;
  errFileOpen   = 2;
  errFileClose  = 3;
  errFileRead   = 4;
  errFileWrite  = 5;
  errFileAttr   = 6;  // Error setting attr.
  errFileFull   = 7;
  errFileGZip   = 8;  // ill. GZip header
  errFileCheck  = 9;  // corrupt copied or packed file
  errFileECrypt = 10; // error encrypting file
  errFileDCrypt = 11; // error decrypting file
  errFileVerify = 12; // verify error
  errLongPath   = 13; // path too long (>260)
  errNotFound   = 14; // source file not found
  errFileTS     = 15; // Error setting timestamp after copy
  errStorage    = 16; // Error copying file details
  errTimeout    = 17; // Timeout error on file transfer
  errStream     = 18; // undefined stream
  errAcl        = 19; // Error copying ACL
  errFileExists = 20; // File already exists
  errFtpRead    = 24; // error reading via FTP
  errFtpWrite   = 25; // error writing via FTP
  errFtpConnect = 26; // connection error on FTP
  errFtpBroken  = 27; // connection reset by server
  errFTPDatConn = 28; // error opening data connection
  errFtpTimeout = 29; // Timeout error on file transfer
  errSignature  = 32; // ill. Zip signature
  errExtract    = 33; // Error extracting file
  errFormat     = 34; // Unsupported file format
  errTmpFile    = 35; // Temp. file could not be renamed

  errUserBreak  = $80; // process stopped bei user
  errAllCodes   = $00FF;

  errError      = $0100;
  errWarning    = $0200;
  errError2     = $0300;
  errAllTypes   = $0F00;

  errCopy       = $1000;
  errGZip       = $2000;
  errGUnzip     = $3000;
  errZip        = $4000;
  errEncrypt    = $5000;
  errDecrypt    = $6000;
  errUnzip      = $7000;
  errAllSources = $F000;
  
function rsErrOpening : string;
function rsErrCreating : string;
function rsErrReading : string;
function rsErrWriting : string;
function rsErrClosing : string;
function rsErrSetAttr : string;
function rsErrTimeStamp : string;
function rsErrAcl : string;
function rsErrDirEntry : string;
function rsErrEndRecord : string;
function rsErrVerify : string;
function rsErrNotFound : string;

// get error message
function GetCopyErrMsg (AError : integer) : string;

implementation

uses SysUtils, GnuGetText;

{ ------------------------------------------------------------------- }
// get error message
function rsErrOpening : string;
begin
  Result:=dgettext('units','Error opening "%s"');
  end;

function rsErrCreating : string;
begin
  Result:=dgettext('units','Error creating "%s"');
  end;

function rsErrReading : string;
begin
  Result:=dgettext('units','Error reading from "%s"');
  end;

function rsErrWriting : string;
begin
  Result:=dgettext('units','Error writing to "%s"');
  end;

function rsErrClosing : string;
begin
  Result:=dgettext('units','Error closing "%s"');
  end;

function rsErrSetAttr : string;
begin
  Result:=dgettext('units','Error setting attributes of "%s"');
  end;

function rsErrTimeStamp : string;
begin
  Result:=dgettext('units','Error setting timestamp of "%s"');
  end;

function rsErrAcl : string;
begin
  Result:=dgettext('units','Error setting permissions of "%s"');
  end;

function rsErrDirEntry : string;
begin
  Result:=dgettext('units','Error writing directory entry to "%s"');
  end;

function rsErrEndRecord : string;
begin
  Result:=dgettext('units','Error writing end record to "%s"');
  end;

function rsErrVerify : string;
begin
  Result:=dgettext('units','Error verifying "%s"');
  end;

function rsErrNotFound : string;
begin
  Result:=dgettext('units','File not found: "%s"');
  end;

function GetCopyErrMsg (AError : integer) : string;
var
  s : string;
begin
  case AError and errAllTypes of
  errError   : s:=dgettext('units','Error: ');
  errError2  : s:=dgettext('units','Error(2): ');
  errWarning : s:=dgettext('units','Warning: ');
  else s:=dgettext('units','Info: ');
    end;
  case AError and errAllCodes of
  errFileCreate : s:=s+dgettext('units','Could not create file');
  errFileOpen   : s:=s+dgettext('units','Could not open file');
  errFileClose  : s:=s+dgettext('units','Could not close file');
  errFileRead   : s:=s+dgettext('units','Could not read from file');
  errFileWrite  : s:=s+dgettext('units','Could not write to file');
  errFileAttr   : s:=s+dgettext('units','File attributes could not be set');
  errFileFull   : s:=s+dgettext('units','Low disk space');
  errFileGZip   : s:=s+dgettext('units','Illegal file header');
  errFileCheck  : s:=s+dgettext('units','Corrupt file');
  errFileECrypt : s:=s+dgettext('units','Encryption failed');
  errFileDCrypt : s:=s+dgettext('units','Decryption failed');
  errFileVerify : s:=s+dgettext('units','Verify failed');
  errLongPath   : s:=s+dgettext('units','Path too long');
  errNotFound   : s:=s+dgettext('units','File not found');
  errFileTS     : s:=s+dgettext('units','Timestamp could not be set');
  errStorage    : s:=s+dgettext('units','File details could not be copied');
  errTimeout    : s:=s+dgettext('units','Timeout error on copying file');
  errStream     : s:=s+dgettext('units','Undefined stream');
  errAcl        : s:=s+dgettext('units','Permissions could not be copied');
  errFileExists : s:=s+dgettext('units','File already exists');
  errFtpRead    : s:=s+dgettext('units','Could not read from FTP');
  errFtpWrite   : s:=s+dgettext('units','Could not write to FTP');
  errFtpConnect : s:=s+dgettext('units','Could not connect to FTP server');
  errFtpBroken  : s:=s+dgettext('units','FTP connection was closed by server');
  errFtpDatConn : s:=s+dgettext('units','Could not open FTP data connection');
  errFtpTimeout : s:=s+dgettext('units','Timeout error on copying file via FTP');
  errSignature  : s:=s+dgettext('units','Illegal zip signature');
  errExtract    : s:=s+dgettext('units','Could not extract file');
  errFormat     : s:=s+dgettext('units','Unsupported file format');
  errTmpFile    : s:=s+dgettext('units','Could not rename temporary file');
  errUserBreak  : s:=s+dgettext('units','Terminated by user');
  else s:=s+Format(dgettext('units','Unknown code (%s)'),[IntToHex(AError,8)]);  // should not happen
    end;
  case AError and errAllSources of
  errCopy       : s:=s+dgettext('units',' (Copy)');
  errGZip       : s:=s+dgettext('units',' (gzip)');
  errGUnzip     : s:=s+dgettext('units',' (gunzip)');
  errZip        : s:=s+dgettext('units',' (Zip)');
  errUnzip      : s:=s+dgettext('units',' (Unzip)');
  errEnCrypt    : s:=s+dgettext('units',' (Encrypt)');
  errDeCrypt    : s:=s+dgettext('units',' (Decrypt)');
    end;
  Result:=s;
  end;

end.

