(* Copy, pack and unpack files using file transfer protocol (FTP)
   optional with encryption

   using Indy 10 routines
   ======================

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   Acknowledgements:
     ZLibEx (Vers. 1.2.3) and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 -   Feb. 2006
         1.1 - May 2006
         2 -   Jul. 2007  - support proxy
         2.1 - Sep. 2007 - support for servers like TOPS-20, see. RFC959
         2.2 - May 2008 - support of MDTM, UTF8 on listing
         2.3 - Aug. 2008 - support SSL/TLS
               Requires: Indy10-Versions of libeay32.dll, ssleay32.dll
                         SSL certificate (created by OpenSSL)
               Added: file match function supporting wildcards
   Vers. 3   - Feb. 2009 : Unicode Version for Delphi 2009
                           ===============================
         3.1 - Dec. 2011 : Compare thread
         3.2 - Jul. 2013 : redesign of TxxFtpThreads

   Required changes in Indy10 library files distributed with Delphi 2009 (Vers. 10.2.3) and newer

   1. Protocols\IdFTP.pas
      - procedure DoAfterLogin; virtual;
      - procedure TIdFTP.SetModTime(const AFileName: String; const ALocalTime: TDateTime);
        ...
          SendCmd('MDTM '+ FTPDateTimeToMDTMD(ALocalTime, False, False) + ' ' + AFileName, [213,253], enDefault); {do not localize}
          // return code 213 is allowed
      - procedure TIdFTP.SetModTimeGMT(const AFileName: String; const AGMTTime: TDateTime);
        ...
          SendCmd('MDTM '+ FTPDateTimeToMDTMD(AGMTTime + OffSetFromUTC, False, False) + ' ' + AFileName, [213,253]); {do not localize}
          // return code 213 is allowed

   2. Protocols\
      // move to public:
      -  property ModifiedDateGMT : TDateTime read FModifiedDateGMT write FModifiedDateGMT; //JR - moved to public 08-12-06

   Changes in Indy10 library files distributed with Delphi 2009

   3. Core\IdTCPConnection.pas:      (in Vers. 10.5.8 genderte Encoding-Routinen, daher nicht erforderlich)
      - default encodings set to enDefault instead of en7bit to support utf8 feature
   4. Protocols\IdMessageClient.pas  (in Vers. 10.5.8 bereits korrigiert)
      - Line 1214 ff
          if LFileName <> '' then begin
            IOHandler.Write(';' + EOL + '        name="' + LFileName + '"'); {do not localize}           // JR 09-11-23
          end;
          IOHandler.WriteLn;
          IOHandler.WriteLn('Content-Transfer-Encoding: ' + LAttachment.ContentTransfer); {do not localize}
          IOHandler.Write('Content-Disposition: ' + LAttachment.ContentDisposition); {do not localize}
          if LFileName <> '' then begin
            IOHandler.Write(';' + EOL + '        filename="' + LFileName + '"'); {do not localize}       // JR 09-11-23
          end;
          IOHandler.WriteLn;

   Aktuell: Indy 10.5.8.0 (4645)
   *)

unit FtpFileTools;

interface

uses Vcl.Forms, Winapi.Windows, System.Classes, System.SysUtils,
  ExtFileTools, ZLibex, AesLib, FtStrings, CBFunctions,
  StringUtils, IdFTP, IdAllFTPListParsers, IdComponent, IdGlobal, IdLogFile;

const
  defTimeout = 60000;   // 60 s
  defConnect = 5;       // default number of trying to connect
  defFtpBufferSize = 64*1024;   // default buffer size for FTP copy operations
  defFtpReplace = '_';
  defFtpKeepAlive = 300;        // time in seconds to send a NOOP on the control
                                // channel during long data transfers
  defLogName = 'FtpLog.txt';

type
  TFtpEncoding = (feASCII,fe8Bit,feUtf8);

  TFtpReplies = record
    Command : string;
    Reply   : integer
    end;
const
  KeepAliveCommands : array[0..2] of TFtpReplies= ((Command: 'NOOP'; Reply : 200),
                                                   (Command: 'PWD'; Reply : 257),
                                                   (Command: 'SYST'; Reply : 215));


type
  TFtpProxyParams = record
    Server,
    Username       : string;
    Password       : AnsiString;
    Port           : integer;
    Mode           : TIdFtpProxyType;
    end;

  TFtpParams = record
    Caption,
    Host,Username,
    Directory,
    InvalidChars        : string;
    Password            : AnsiString;
    Port,
    SecureMode,             // 0 = off, 1 = explict, 2 = required, 4 = implicit
    ConnTimeout,
    ConnAttempts,
    ResponseTimeOut,
    KeepAlive,
    KeepAliveCmd        : integer;   // 0 = NOOP, 1 = PWD, 2 = SYST
    Quotes,
    Passive,
    UseHost,
    WriteLog,
    ForceUtf8           : boolean;
    CaseMode            : TTextChange;
    Proxy               : TFtpProxyParams
    end;

  TDirectoryInfo = class(TObject)
    FFileName : string;
    FSize     : Int64;
    FModifiedDate    : TDateTime;
    FModifiedDateGMT : TDateTime;
    constructor Create (const AFilename : string; ASize : int64; AMd,AMdG : TDateTime);
    end;

  TDirectoryList = class(TStringlist)
    destructor Destroy; override;
    procedure Clear; override;
    end;

  TGetPassword = procedure(var APassword : AnsiString) of object;

  TExtFtp = class (TIdFTP)
  private
    FList  : TStringList;
    CMode  : TTextChange;
    ValidFileList,
    fQuotes,
    fTLS,
    fUTF8  : boolean;
    fDir,
    fCurrentDir,
    FChangeDirResponse,
    fRepl    : string;
    fCertPwd : AnsiString;
    fGetSSLPassword : TGetPassword;
    fEncoding : TFtpEncoding;
    fIdLogFile : TIdLogFile;
    fForceUtf8 : boolean;
    fRoot,fCert,fKey,
    fTmpDir,fIdError : string;
    DCount : integer;            // count "ChangeDir" in "ForceDir"
    procedure InitIOHandler;
    procedure SetTmpDir (ADir : string);
    function GetLastResponseCode : integer;
    function GetLastResponseString : string;
    function GetEncoding : TFtpEncoding;
    procedure SetEncoding (Value : TFtpEncoding);
    function GetLogName : string;
    procedure SetLogName (Value : string);
    procedure SetWriteLog (Value : boolean);
  protected
    procedure DoAfterLogin; override;
    procedure SetTLSMode(Value : integer);
    procedure GetCertPassword(var Password : AnsiString);
  public
    constructor Create (AOwner : TComponent; const Path,Root,Cert,Key : string; Pwd : AnsiString); overload;
    constructor Create (AOwner : TComponent); overload;
    destructor Destroy; override;
    function ModifyName(const FName : string) : string;
    function RetrieveName(const FName : string) : string;
    function ReProcessFilename (const s : string) : string;
    function ProcessFilename (const s : string) : string;
    function GetFileList (FileList : TStringList) : boolean; overload;
    function GetFileList : boolean; overload;
    function FtpFileExists (const FName : string) : boolean;
    function DeleteSubDirFile (const ADir,FName : string) : integer;
    function DeleteFile (const FName : string) : boolean;
    function DeleteMatchingFiles (const AMask : string) : integer;
    function RenameFile (const OldName,NewName : string) : boolean;
    function GetSize (FileList : TStringList; const FName : string) : int64; overload;
    function GetSize (const FName : string) : int64; overload;
    function CheckSize(const FName : string; ASize : int64) : boolean;
    function GetLocalFileTime (const FName : string) : TFileTime;
    function GetModTime(const AFileName: String; AsGMT: Boolean): TDateTime;
    function GetTimeStamp (FileList : TStringList; const FName : string; AsGMT : Boolean = false) : TDateTime; overload;
    function GetTimeStamp (const FName : string; AsGMT : Boolean = false) : TDateTime; overload;
    function CheckTimeStamp (const FName : string; tm : TDateTime;
                             Delta : integer; AsGMT : Boolean = false) : integer;
    function SetTimeStamp(const FName : string; ATime : TDateTime; UseGMT : Boolean = false) : boolean;
    function SetTimeStampFromFile(const ASourceFile,FName : string;
                                  CheckTime : boolean = true; UseGMT : Boolean = false) : boolean;
    function GetDirList(DirList : TDirectoryList) : boolean;
    function FtpDirExists (const DName : string) : boolean;
    procedure FtpChangeDir (DName: string);
    function ChangeOneDirUp : boolean;
    function ChangeToDir(DName: string) : boolean;
    function ChangeToNewDir(DName: string) : integer;
    function ForceDir(DName: string) : boolean;
    function ForceDirUp : boolean;
    function SetRemoteDir (const Dir : string) : boolean;
    function DeleteDir(DName: string) : boolean;
    function GetCurrentDir : string;
    function DeleteTree(DName : string; var DCount : integer) : boolean;
    function DirFiles(DName : string) : integer;
    function DoConnect (NRepeat : integer = 1) : boolean;
    function ReConnect(const Dir : string; check : boolean = true) : boolean;
    function Connected : boolean; override;
    function CopyFile (const ASourceFile,ADestFile : string) : integer;
    function Flush(ASource: TStream; const ADestFile: string; AAppend: boolean) : integer;
    function Fill(const ASourceFile : string; ADest : TStream) : integer;
    function GzFileInfo (Filename: string; var FileInfo : TGzFileInfo) : integer;
    property CanUseTls : boolean read fTls;
    property CaseMode : TTextChange read CMode write CMode;
    property ChangeDirResponse : string read FChangeDirResponse;
    property CurrentDir : string read fCurrentDir;
    property Encoding : TFtpEncoding read GetEncoding write SetEncoding;
    property FileList : TStringList read FList;
    property ForceUtf8 : boolean read fForceUtf8 write fForceUtf8;
    property IndyError : string read fIdError;
    property LastResponseCode : integer read GetLastResponseCode;
    property LastResponseString : string read GetLastResponseString;
    property RemoteDir : string read fDir;
    property ReplaceChars : string read fRepl write fRepl;
    property UseQuotes : boolean read fQuotes write fQuotes;
    property SecureTransfer : integer write SetTLSMode;
    property TempDir : string read fTmpDir write SetTmpDir;
    property Logname : string read GetLogName write SetLogName;
    property WriteLogFile : boolean write SetWriteLog;
    property OnSSLPassword: TGetPassword read fGetSSLPassword write fGetSSLPassword;
    end;

  // Write file via FTP with opt. encryption
  TWriteFtpThread = class (TGZipThread)
  private
    FFtp      : TExtFtp;
    FCompress,
    FZipCompress,
    FVerify   : boolean;
    FDestDir  : string;
    FFileSize : int64;
  protected
    function WriteFile : integer;
    function VerifyFile (fs : TStream) : integer;
    procedure Execute; override;
    procedure EndThread (Error : integer); override;
    procedure ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
  public
  // ASourceName  = source file,  ADestName =  dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (AFtp : TExtFTP; const ASourceName,ADestDir,ADestName : string;
                        AEncryption : TEncryption; AOptions : TCopyOptions;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; override;
    property DestSize : int64 read FFileSize;
    end;

  // Read file via FTP with opt. encryption
  TReadFtpThread = class (TGUnZipThread)
  private
    FFtp      : TExtFtp;
    FZipCompress,
    FCompress : boolean;
    FFileSize : int64;
  protected
    function ReadFile : integer;
    procedure Execute; override;
    procedure EndThread (Error : integer); override;
    procedure ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
    property DestSize : int64 read FFileSize;
  public
  // ASourceName  = source file,  ADestName =  dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (AFtp : TExtFTP; const ASourceName,ADestName : string; AFileSize : int64;
                        AEncryption : TEncryption; ACompress,AZipCompress,ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; override;
    end;

  TCompareFtpThread = class (TReadFtpThread)
  protected
    function CompareFile : integer;
    procedure Execute; override;
  public
    property DestSize;
    end;

// Bei Verzeichnisangaben Slash ergnzen
function IncludeTrailingSlash (const Dir : string) : string;
function IncludeLeadingSlash (const Dir : string) : string;
function RemoveLeadingSlash (const Dir : string) : string;
function RemoveTrailingSlash (const Dir : string) : string;

// Den letzten Teilnamen extrahieren (siehe auch: ExtractFileName)
function ExtractLastName(const Delimiters,FileName: string): string;
function ExtractPath(const APath : string) : string;

// Unix-Pfad zusammensetzen
function ExpandUnixPath(const APath,AName : string) : string;

// Relativen Unix-Pfad extrahieren
function GetUnixRelativePath(const RootPath,Path : string) : string;

implementation

uses IdFTPList, IdReplyRFC, IdReply, IdFTPCommon, IdSSLOpenSSL, IdIOHandler, IdStack,
  IdExplicitTLSClientServerBase, IdGlobalProtocols, IdException, IdExceptionCore,
  System.DateUtils, System.Masks, FileUtils, WinApiUtils, UnitStrings, UnicodeUtils,
  WinUtils, StrUtils;

{ ------------------------------------------------------------------- }
(* Bei Verzeichnisangaben Slash ergnzen *)
function IncludeTrailingSlash (const Dir : string) : string;
begin
  if (length(Dir)=0) or (Dir[length(Dir)]='/') then Result:=Dir
  else Result:=Dir+'/';
  end;

function IncludeLeadingSlash (const Dir : string) : string;
begin
  if (length(Dir)=0) or (Dir[1]<>'/') then Result:='/'+Dir
  else Result:=Dir;
  end;

function RemoveLeadingSlash (const Dir : string) : string;
begin
  Result:=Dir;
  if (length(Result)>0) and (Result[1]='/') then Delete(Result,1,1);
  end;

function RemoveTrailingSlash (const Dir : string) : string;
begin
  Result:=Dir;
  if (length(Result)>0) and (Result[length(Result)]='/') then Delete(Result,length(Result),1);
  end;

// Den letzten Teilnamen extrahieren (siehe auch: ExtractFileName)
function ExtractLastName(const Delimiters,FileName: string): string;
var
  I: Integer;
begin
  I:=LastDelimiter(Delimiters, FileName);
  Result:=Copy(FileName,I+1,MaxInt);
  end;

function ExtractPath(const APath : string) : string;
begin
  Result:=IndyGetFilePath(APath);
  end;

// Unix-Pfad zusammensetzen
function ExpandUnixPath(const APath,AName : string) : string;
begin
  if length(APath)>0 then Result:=APath+'/'+AName
  else Result:=AName;
  end;

// Relativen Unix-Pfad extrahieren
function GetUnixRelativePath(const RootPath,Path : string) : string;
begin
  Result:=DOSPathToUnixPath(Path);
  if AnsiStartsText(DOSPathToUnixPath(RootPath),Result) then begin
    Delete(Result,1,length(RootPath));
    end
  else Result:='';
  end;

{ ------------------------------------------------------------------- }
constructor TDirectoryInfo.Create (const AFilename : string; ASize : int64; AMd,AMdG : TDateTime);
begin
  inherited Create;
  FFilename:=AFilename; FSize:=ASize;
  FModifiedDate:=AMd; FModifiedDateGMT:=AMdG;
  end;

Destructor TDirectoryList.Destroy;
begin
  Clear;
  inherited Destroy;
  end;

procedure TDirectoryList.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do if assigned(Objects[i]) then begin
    Objects[i].Free; Objects[i]:=nil;
    end;
  inherited Clear;
  end;

{ ------------------------------------------------------------------- }
constructor TExtFtp.Create (AOwner : TComponent; const Path,Root,Cert,Key : string; Pwd : AnsiString);
begin
  inherited Create(AOwner);
  UseMLIS:=true; ValidFileList:=false;
  Passive:=true; fUTF8:=false; fQuotes:=false; fCertPwd:=''; UseHost:=true;
  TransferType:=ftBinary;    // in den verw. Indy-Routinen ist ASCII per default eingestellt
  FList:=TStringList.Create;
  CMode:=tcNone; fIdError:='';
  fDir:=''; fCurrentDir:='';
  fIdLogFile:=TIdLogFile.Create(nil);          //create the log file
  with fidLogFile do begin
    FileName:=SetDirName(fTmpDir)+defLogName;
    end;
// if openSSL libraries and certificates are available, use SSL for TLS support
  fTLS:=(length(Root)>0) and (length(Cert)>0) and (length(Key)>0);
  if fTLS then begin
    try
      TIdSSLContext.Create.Free;
      fTLS:=true;
    except
      fTLS:=false;
      end;
    if fTLS then begin
      fRoot:=Path+Root; fCert:=Path+Cert; fKey:=Path+Key;
      fTLS:=FileExists(fRoot) and FileExists(fCert) and FileExists(fKey);
      end;
    end;
  fCertPwd:=Pwd;
  InitIOHandler;
  UseTLS:=utNoTLSSupport;
  end;

procedure TExtFtp.InitIOHandler;
begin
  if IOHandler<>nil then Exit;
  if fTLS then begin
    IOHandler:=TIdSSLIOHandlerSocketOpenSSL.Create(self);
    with (IOHandler as TIdSSLIOHandlerSocketOpenSSL) do begin
      with SSLOptions do begin
        RootCertFile:=fRoot; CertFile:=fCert; KeyFile:=fKey;
        end;
      OnGetPassword:=GetCertPassword;
      end;
    end
  else IOHandler:=TIdIOHandler.MakeDefaultIOHandler(self);
  IOHandler.Intercept:=fIdLogFile;
  ManagedIOHandler:=True;
  end;

constructor TExtFtp.Create (AOwner : TComponent);
begin
  Create(AOwner,'','','','','');
  end;

destructor TExtFtp.Destroy;
begin
  fIdLogFile.Free;
  FreeListObjects(FList);
  FList.Free;
  inherited Destroy;
  end;

procedure TExtFtp.DoAfterLogin;
begin
  fUTF8:=IsExtSupported('UTF8');
  GetEncoding;
  if Assigned(FOnAfterClientLogin) then begin
    OnAfterClientLogin(Self);
    end;
  end;

procedure TExtFtp.SetTmpDir (ADir : string);
begin
  fTmpDir:=SetDirName(ADir);
  end;

function TExtFtp.GetLastResponseCode : integer;
begin
  Result:=LastCmdResult.NumericCode;
  end;

function TExtFtp.GetLastResponseString : string;
begin
  Result:=LastCmdResult.Code;
  if length(fIdError)>0 then Result:=Result+' ('+fIdError+')';
  end;

function TExtFtp.GetLogName : string;
begin
  Result:=fIdLogFile.FileName;
  end;

procedure TExtFtp.SetLogName (Value : string);
begin
  fIdLogFile.FileName:=Value;
  end;

procedure TExtFtp.SetWriteLog (Value : boolean);
begin
  fIdLogFile.Active:=Value;
  end;

function TExtFtp.GetEncoding : TFtpEncoding;
begin
  with IOHandler do begin
    if DefStringEncoding=IndyUTF8Encoding then Result:=feUtf8
    else if DefStringEncoding=Indy8BitEncoding then Result:=fe8Bit
    else Result:=feASCII;
    end;
  FEncoding:=Result;
  end;

procedure TExtFtp.SetEncoding (Value : TFtpEncoding);
begin
  if Value=FEncoding then Exit;
  if Value=feUtf8 then begin
    if not SendCmd('OPTS UTF8 ON') in [200, 220] then begin {do not localize}
    // trying draft-ietf-ftpext-utf-8-option-00.txt next...
      if SendCmd('OPTS UTF-8 NLST') <> 200 then Exit;
      end;
    IOHandler.DefStringEncoding:=IndyUTF8Encoding;
    end
  else begin
    SendCmd('OPTS UTF8 OFF');
    if Value=fe8Bit then IOHandler.DefStringEncoding:=Indy8BitEncoding
    else IOHandler.DefStringEncoding:=IndyASCIIEncoding;
    end;
  FEncoding:=Value;
  end;

procedure TExtFtp.SetTLSMode(Value : integer);
begin
  if fTLS and (Value>0) then begin
    if Value=1 then UseTLS:=utUseExplicitTLS
    else if Value=2 then UseTLS:=utUseRequireTLS
    else UseTLS:=utUseImplicitTLS;
    DataPortProtection:=ftpdpsPrivate;
    end
  else begin
    UseTLS:=utNoTLSSupport;
    DataPortProtection:=ftpdpsClear;
    end;
  end;

procedure TExtFtp.GetCertPassword(var Password : AnsiString);
begin
  if Assigned(fGetSSLPassword) then fGetSSLPassword(Password)  // prompt for password
  else Password:=fCertPwd;
  end;

function TExtFtp.ModifyName(const FName : string) : string;
begin
  // replace invalid characters and change case
  Result:=TextChangeCase(SpecCharToValue(FName,fRepl),CMode);
  end;

function TExtFtp.RetrieveName(const FName : string) : string;
begin
  if length(fRepl)=0 then Result:=FName else Result:=ValueToSpecChar(FName);
  end;

// Die Konvertierung auf UTF8 wird in den Indy-Routinen vorgenommen, dazu musste
// in IdTCPConnection berall en7Bit durch enDefault ersetzt werden
function TExtFtp.ReProcessFilename (const s : string) : string;
begin
//  if fUTF8 then Result:=Utf8ToUnicodeString(s) else
    Result:=s;
  end;

function TExtFtp.ProcessFilename (const s : string) : String;
begin
//  if fUTF8 then Result:=UTF8Encode(s) else
  if fQuotes and (pos(' ',s)>0) then Result:=AnsiQuotedStr(s,StringUtils.Quote)
  else Result:=s;
  end;

function TExtFtp.GetFileList (FileList : TStringList) : boolean;
var
  i,n : integer;
begin
  Result:=false; n:=3;
  FileList.Clear;
  repeat
    try
      List (nil);
      with DirectoryListing do begin
        // Dateien
        for i:=0 to Count-1 do with Items[i] do if ItemType<>ditDirectory then
          FileList.AddObject(RetrieveName(ReProcessFilename(Filename)),Items[i]);
        end;
      Result:=true;
    except
      on E:EIdReadTimeout do n:=1;
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
        n:=1;
        end;
      else n:=1;
      end;
    dec(n);
    until Result or (n=0);
  end;

function TExtFtp.GetFileList : boolean;
begin
  if ValidFileList then Result:=true
  else begin
    Result:=GetFileList(FList);
    if Result then ValidFileList:=true;
    end;
  end;

// die Objekte mssen von der aufrufenden Funktion entfernt werden
function TExtFtp.GetDirList (DirList : TDirectoryList) : boolean;
var
  i   : integer;
begin
  Result:=true;
  DirList.Clear;
  try
    List (nil);
    with DirectoryListing do begin
      // Verzeichnisse
      for i:=0 to Count-1 do with Items[i] do
        if (ItemType=ditDirectory) and NotSpecialDir(Filename) then
          DirList.AddObject(RetrieveName(ReProcessFilename(Filename)),
            TDirectoryInfo.Create(Filename,Size,ModifiedDate,ModifiedDateGMT));
      end;
  except
    Result:=false;
    end;
  end;

function TExtFtp.FtpDirExists (const DName : string) : boolean;
var
  dl : TDirectoryList;
begin
  Result:=false;
  dl:=TDirectoryList.Create;
  if GetDirList(dl) then Result:=dl.IndexOf(DName)>=0;
  dl.Free;
  end;

function TExtFtp.FtpFileExists (const FName : string) : boolean;
begin
  Result:=false;
  if GetFileList then Result:=FileList.IndexOf(FName)>=0;
  end;

// Delete file in subfolder
// Result = 0 : failed
//        = 1 : ok
//        = 2 : not found
function TExtFtp.DeleteSubDirFile (const ADir,FName : string) : integer;
var
  fl : TStringList;
begin
  Result:=0;
  if ChangeToDir(ADir) then begin
    fl:=TStringList.Create;
    if GetFileList(fl) then begin
      if fl.IndexOf(FName)>=0 then begin
        if DeleteFile(FName) then Result:=1;
        end
      else Result:=2;
      end;
    fl.Free;
    ChangeOneDirUp;
    end;
  end;

function TExtFtp.DeleteFile (const FName : string) : boolean;
var
  n   : integer;
begin
  Result:=false; n:=3;
  ValidFileList:=false;
  repeat
    try
      Delete(ProcessFilename(ModifyName(FName)));
      Result:=true;
    except
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
        n:=1;
        end;
      else n:=1;
      end;
    dec(n);
    until Result or (n=0);
  end;

function TExtFtp.RenameFile (const OldName,NewName : string) : boolean;
var
  n   : integer;
begin
  Result:=false; n:=3;
  ValidFileList:=false;
  repeat
    try
      Rename(ProcessFilename(ModifyName(OldName)),ProcessFilename(ModifyName(NewName)));
      Result:=true;
    except
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
//        n:=1;
        end;
      else n:=1;
      end;
    dec(n);
    until Result or (n=0);
  end;

function TExtFtp.DeleteMatchingFiles (const AMask : string) : integer;
var
  i : integer;
begin
  Result:=0;
  with FileList do for i:=0 to Count-1 do if MatchesMask(Strings[i],AMask) then begin
    DeleteFile(ProcessFilename(Strings[i]));
    inc(Result);
    end;
  end;

function TExtFtp.GetCurrentDir : string;
var
  n  : integer;
  ok : boolean;
begin
  Result:=''; ok:=false; n:=3;
  repeat
    try
      Result:=RetrieveName(ReProcessFilename(RetrieveCurrentDir));
      ok:=true;
    except
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
        n:=1;
        end;
      else n:=1;
      end;
    dec(n);
    until ok or (n=0);
  end;

// erfordert vorheriges GetFileList fr das aktuelle Verzeichnis
function TExtFtp.GetSize (const FName : string) : int64;
begin
  Result:=GetSize (FList,FName);
  end;

function TExtFtp.GetSize (FileList : TStringList; const FName : string) : int64;
var
  n : integer;
begin
  n:=FileList.IndexOf(FName);
  if n>=0 then Result:=(FileList.Objects[n] as TIdFTPListItem).Size
  else Result:=0;
  end;

function TExtFtp.CheckSize(const FName : string; ASize : int64) : boolean;
begin
  Result:=GetSize(FName)<>ASize;
  end;

// Replacement for "FileDate"
// Uses the same dependances on features as SetModTime
function TExtFtp.GetModTime(const AFileName: String; AsGMT: Boolean): TDateTime;
var
  LBuf : String;
begin
  if SendCmd('MDTM ' + AFileName) = 213 then begin {do not localize}
    LBuf := LastCmdResult.Text[0];
    LBuf := Trim(LBuf);
    if AsGMT then Result := FTPMLSToGMTDateTime(LBuf)
    else begin
      if IsExtSupported('MFMT') then Result:=FTPMLSToLocalDateTime(LBuf)
      //syntax 1 + 2
      else if (IndexOfFeatLine('MDTM YYYYMMDDHHMMSS[+-TZ];filename') > 0) or
         (IndexOfFeatLine('MDTM YYYYMMDDHHMMSS[+-TZ] filename') > 0) or IsOldServU or IsBPFTP then
        Result := FTPMLSToLocalDateTime(LBuf)
      //syntax 3 - MDTM [local timestamp] Filename
      else if TZInfo.GMTOffsetAvailable then Result:=FTPMLSToGMTDateTime(LBuf)-TZInfo.GMTOffset
      else Result:=FTPMLSToGMTDateTime(LBuf);
      end;
    end
  else Result := 0;
  end;

// erfordert vorheriges GetFileList fr das aktuelle Verzeichnis
// Wenn MLST untersttzt wird kommt der Zeitstempel von dort,
// andernfalls wird nicht mehr FileDate, sondern GetModTime (s.o.) verwendet
function TExtFtp.GetTimeStamp (const FName : string; AsGMT : Boolean = false) : TDateTime;
begin
  Result:=GetTimeStamp (FList,FName,AsGMT);
  end;

function TExtFtp.GetTimeStamp (FileList : TStringList; const FName : string; AsGMT : Boolean = false) : TDateTime;
var
  n : integer;
begin
  n:=FileList.IndexOf(FName);
  if n>=0 then begin
    if AsGMT then Result:=(FileList.Objects[n] as TIdFTPListItem).ModifiedDateGMT
    else Result:=(FileList.Objects[n] as TIdFTPListItem).ModifiedDate;
    if not ((FCapabilities.Count>0) and FCanUseMLS) or (Result<1) then begin
      // get local time from server because LIST may return time stamps
      // with only date information (time = 0)
      Result:=GetModTime(ProcessFilename(ModifyName(FName)),AsGMT);  // change to local time
      end
    end
  else Result:=0;
  end;

// Compare timestamps as local or UTC time, Delta in s
// Result = 0 : FName not found
//          1 : timestamp of FName < tm - Delta
//          2 : tm - Delta <= timestamp of FName <= tm + Delta
//          3 : tm + Delta < timestamp of FName
function TExtFtp.CheckTimeStamp (const FName : string; tm : TDateTime;
                                 Delta : integer; AsGMT : Boolean = false) : integer;
var
  md : TDateTime;
begin
  md:=GetTimeStamp(FName,AsGMT);
  if md>1 then begin
    if md>=IncSecond(tm,-Delta) then begin
      if md<=IncSecond(tm,Delta) then Result:=2 else Result:=3
      end
    else Result:=1;
    end
  else Result:=0;
  end;

function TExtFtp.GetLocalFileTime (const FName : string) : TFileTime;
begin
  Result:=DateTimeToFileTime(GetTimeStamp(FName,false));
  end;

function TExtFtp.SetTimeStamp(const FName : string; ATime : TDateTime; UseGMT : Boolean = false) : boolean;
begin
  Result:=true;
  if (FCapabilities.Count>0) and FCanUseMLS then begin
    try
      // filetime from source is GMT
      if UseGMT then SetModTimeGMT(ProcessFilename(ModifyName(FName)),ATime)
      else SetModTime(ProcessFilename(ModifyName(FName)),ATime);
    except
      Result:=false
      end;
    end
  else begin
    try
      SetModTime(ProcessFilename(ModifyName(FName)),ATime);
    except
      Result:=false
      end;
    end;
  end;

function TExtFtp.SetTimeStampFromFile(const ASourceFile,FName : string;
                     CheckTime : boolean = true; UseGMT : Boolean = false) : boolean;
var
  dt : TDateTime;
begin
  if UseGmt then dt:=FileTimeToDateTime(GetFileLastWriteTime(ASourceFile))
  else dt:=FileAgeUNC(ASourceFile,0);
  if CheckTime and (dt>Now+1) then dt:=Now;
  Result:=SetTimeStamp(FName,dt,UseGMT);
  end;

procedure TExtFtp.FtpChangeDir (DName: string);
begin
  ValidFileList:=false;
  DName:=ModifyName(ProcessFilename(DName));
  if AnsiStartsText(fDir,DName) then
    DName:=fDir+copy(DName,length(fDir)+1,length(DName));
  inherited ChangeDir(DName);
  end;

// ChangeDirUp with boolean result
function TExtFtp.ChangeOneDirUp : boolean;
begin
  Result:=true;
  ValidFileList:=false;
  try ChangeDirUp; except Result:=false; end;
  with LastCmdResult do FChangeDirResponse:=Code+' '+Text[0];
  fCurrentDir:=GetCurrentDir;
  end;

// ChangeDir with boolean result
function TExtFtp.ChangeToDir(DName: string) : boolean;
begin
  Result:=true;
  if length(DName)=0 then Exit;
  try FtpChangeDir(DName); except Result:=false; end;
  with LastCmdResult do FChangeDirResponse:=Code+' '+Text[0];
  fCurrentDir:=GetCurrentDir;
  end;

// Change to dir and create if not exists
// return : -1 = error, 0 = ok exists, 1 = ok is new
function TExtFtp.ChangeToNewDir(DName: string) : integer;
var
  s : string;
  n : integer;
begin
  Result:=-1;
  try
    FtpChangeDir(DName);
    with LastCmdResult do FChangeDirResponse:=Code+' '+Text[0];
    Result:=0
  except
    try
      MakeDir(ProcessFilename(ModifyName(DName)));
      Result:=1;
      with LastCmdResult do begin
        FChangeDirResponse:=Code+' '+Text[0];
        if NumericCode=257 then begin
          s:=Text[0];
          n:=LastDelimiter('"',s);
          if n>0 then begin
            s:=AnsiDequotedStr(copy(s,1,n),'"');
            if Pos(#$fffd,s)>0 then s:=DName;       // wird bei proftp zurckgegeben
            end
          else s:=DName;
          end
        else s:=DName;
        end;
      try FtpChangeDir(s); except Result:=-1; end;
    except
      with LastCmdResult do begin
        FChangeDirResponse:=Code+' '+Text[0];
      // Accept "250" as legal response (Some FTP servers need this)
      // Accept "200" as legal response (DSL-EasyBox 803 FTP Server)
        if (NumericCode=250) or (NumericCode=200) then begin
          Result:=1;
          try FtpChangeDir(DName); except Result:=-1; end;
          end;
        end;
      end;
    end;
  fCurrentDir:=GetCurrentDir;
  end;

// Create subdirectory(ies)
// Result = true if DName exists or successfully created
// count relative level changes (DCount) for use in "ForceDirUp"
function TExtFtp.ForceDir(DName: string) : boolean;
var
  ok,root : boolean;
  d  : string;
  n  : integer;
begin
  Result:=true; root:=false;
  DCount:=0;
  if length(DName)>0 then begin
    DName:=ModifyName(DName);
    ok:=true;
    try
      FtpChangeDir(ProcessFilename(DName));
      DCount:=CountSubStr('/',DName)+1;
    except
      ok:=false;
      end;
    if not ok then begin
      if (length(fDir)>0) and AnsiStartsText(fDir,DName) then begin
        try
          FtpChangeDir (fDir);
          if length(fDir)=1 then n:=1 else n:=2;
          DName:=copy(DName,length(fDir)+n,length(DName));
          ok:=true; root:=true;
        except
          end;
        end
      else ok:=true;
      while ok and (length(DName)>0) do begin
        d:=ReadNxtStr(DName,'/');
        if ChangeToNewDir(d)>=0 then inc(DCount)
        else ok:=false;
        end;
      if root then DCount:=0; // was abs. path
      Result:=ok;
      end
    end;
  fCurrentDir:=GetCurrentDir;
  end;

// Go up all subdirectories create by last "ForceDir"
function TExtFtp.ForceDirUp: boolean;
var
  i : integer;
begin
  Result:=false;
  ValidFileList:=false;
  for i:=1 to DCount do begin
    try ChangeDirUp; except exit; end;
    end;
  DCount:=0;
  fCurrentDir:=GetCurrentDir;
  Result:=true;
  end;

function TExtFtp.SetRemoteDir (const Dir : string) : boolean;
begin
  Result:=ForceDir(Dir);
  fCurrentDir:=GetCurrentDir;
  end;

function TExtFtp.DeleteDir (DName : string) : boolean;
begin
  Result:=true;
  ValidFileList:=false;
  try
    RemoveDir(ProcessFilename(ModifyName(DName)));
  except
    Result:=false;
    end;
  end;

// delete tree including all files
function TExtFtp.DeleteTree (DName : string; var DCount : integer) : boolean;
var
  i  : integer;
  DList : TDirectoryList;
  sc : string;
begin
  Result:=false;
  ValidFileList:=false;
  sc:=GetCurrentDir;
  if length(DName)>0 then begin
    if not ChangeToDir(DName) then begin
      FtpChangeDir(sc);
      Exit;
      end;
    end;
  List (nil);
  // Delete files
  with DirectoryListing do for i:=0 to Count-1 do with Items[i] do
      if (ItemType<>ditDirectory) then begin
    try
      self.Delete(ProcessFilename(Filename));
      inc(DCount);
    except
      end;
    Application.ProcessMessages;
    end;
  DList:=TDirectoryList.Create;
  if GetDirList(DList) then begin
    i:=0;
    Result:=true;
    with DList do while (i<Count) and Result do begin
      Result:=DeleteTree(Strings[i],DCount);
      inc(i);
      end;
    if length(DName)>0 then begin
      try ChangeDirUp; except; end;
      if Result then begin
        try Result:=DeleteDir(DName); except; end;
        end;
      end;
    end;
  DList.Free;
  end;

function TExtFtp.DirFiles(DName : string) : integer;
var
  DList : TDirectoryList;
  i     : integer;
begin
  DName:=ModifyName(DName);
  Result:=0;
  try
    if length(DName)>0 then begin
      try FtpChangeDir(ProcessFilename(DName)); except; end;
      end;
    List (nil);
    // Count files
    with DirectoryListing do for i:=0 to Count-1 do with Items[i] do
      if (ItemType<>ditDirectory) then inc(Result);
  except
    if length(DName)>0 then begin
      try FtpChangeDir(ProcessFilename(DName)); except; end;
      end;
    Exit;
    end;
  DList:=TDirectoryList.Create;
  if GetDirList(DList) then begin
    with DList do for i:=0 to Count-1 do Result:=Result+DirFiles(Strings[i]);
    if length(DName)>0 then begin
      try ChangeDirUp; except; end;
      end;
    end;
  DList.Free;
  end;

// Connect to FTP,
// if server does not support SSL on utUseExplicitTLS reset DataPortProtection
function TExtFtp.DoConnect (NRepeat : integer = 1) : boolean;
begin
  fIdError:='';
  ValidFileList:=false;
  repeat
    try
      Connect;
      Result:=true;
      fDir:=GetCurrentDir;
      if (UseTLS=utUseExplicitTLS) and not FUsingSFTP then begin // TLS not available
        DataPortProtection:=ftpdpsClear;
        end;
      if ForceUtf8 then Encoding:=feUtf8;
    except
      Sleep(1000);
      Result:=false;
      end;
    dec(NRepeat);
    if NRepeat>0 then InitIOHandler;
    until Result or (NRepeat<=0);
  end;

// Check connection and try to reconnect if connection was cancelled
function TExtFtp.ReConnect(const Dir : string; Check : boolean = true) : boolean;
var
  sd : string;
begin
  if Check then begin
    try
      sd:=RetrieveCurrentDir;  // check if still connected
      if not AnsiSameText(sd,ProcessFilename(ModifyName(Dir))) then ChangeToDir(Dir);
      Result:=true;
      Exit;
    except
      // ClosedGracefully
      try Disconnect except end;
      end;
    end;
  Result:=DoConnect(3);
  if Result then begin
    ChangeToDir(Dir);
    GetFileList;
    end;
  end;

// Replaces original function to check if peer has closed connection
function TExtFtp.Connected : boolean;
begin
  try
    Result:=inherited Connected;
  except
    Result:=false;
    end;
  end;

// Copy file
function TExtFtp.CopyFile (const ASourceFile,ADestFile : string) : integer;
var
  fs : TFileStream;
begin
  Result:=errOK;
  if not FileExistsUNC(ASourceFile) then Result:=errError+errFileOpen;
  if (Result=errOK) and (length(ADestFile)=0) then Result:=errError+errFileCreate;
  if Result=errOK then begin
    try
      fs:=TFileStream.Create(FilenameToUNC(ASourceFile),fmOpenRead+fmShareDenyNone);
    except
      Result:=errError+errFileOpen;
      end;
    end;
  if Result=errOK then Result:=Flush(fs,ADestFile,false);
  end;

// Write stream
function TExtFtp.Flush(ASource: TStream; const ADestFile: string; AAppend: boolean) : integer;
var
  n   : integer;
begin
  if not ReConnect(fCurrentDir) then begin
    Result:=errError+errFtpBroken; exit;
    end;
  n:=3;
  repeat
    try
      Put(ASource,ProcessFilename(ModifyName(ADestFile)),AAppend);
      if LastCmdResult.NumericCode<400 then Result:=errOK    // siehe RFC 640
      else Result:=errError+errFtpWrite;
    except
      on E:EIdReadTimeout do begin
        n:=1;
        Result:=errError+errFtpTimeout;
        end;
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        Result:=errError+errFtpBroken;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        Result:=errError+errFtpBroken;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
        n:=1;
        if LastCmdResult.NumericCode=425 then Result:=errError+errFtpDatConn
        else Result:=errError+errFtpWrite;
        end;
      else begin
        n:=1;
        Result:=errError+errFtpWrite;
        end;
      end;
    dec(n);
    until (Result=errOK) or (n=0);
  if (Result=errOK) and not ReConnect(fCurrentDir) then Result:=errError+errFtpBroken;
  end;

// Stream lesen
function TExtFtp.Fill(const ASourceFile : string; ADest : TStream) : integer;
var
  n   : integer;
begin
  if not ReConnect(fCurrentDir) then begin
    Result:=errError+errFtpBroken; exit;
    end;
  n:=3;
  repeat
    try
      Get(ProcessFilename(ModifyName(ASourceFile)),ADest,false);
      ADest.Seek(0,soFromBeginning);
      Result:=errOK;
    except
      on E:EIdSocketError do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        Result:=errError+errFtpBroken;
        end;
      on E:EIdConnClosedGracefully do begin
        fIdError:=E.Message;
        try ReConnect(fCurrentDir,false); except end;
        Result:=errError+errFtpBroken;
        end;
      on E:EIdException do begin
        fIdError:=E.Message;
        n:=1;
        Result:=errError+errFtpRead;
        end;
      else begin
        n:=1;
        Result:=errError+errFtpRead;
        end;
      end;
    dec(n);
    until (Result=errOK) or (n=0);
  if (Result=errOK) and not ReConnect(fCurrentDir) then Result:=errError+errFtpBroken;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from gz-file *)
function TExtFtp.GzFileInfo (Filename: string; var FileInfo : TGzFileInfo) : integer;
var
  sGz : TFileStream;
  tzn : string;
begin
  Result:=errError+errFileRead;
  try
    tzn:=TempDir+tmpGzipName;
    sGz:=TFileStream.Create(tzn,fmCreate); // write to disk
    if (Fill(Filename,sGz)=errOK) then begin
      sGz.Seek(0,soFromBeginning);
      if GetGzInfo(sGz,FileInfo) then result:=errOK;
      end
  finally
    sGz.Free;
    DeleteFile(tzn);
    end;
  end;

{ TWriteFtpThread ------------------------------------------------------------- }
  // Write file via FTP with opt. encryption
constructor TWriteFtpThread.Create (AFtp : TExtFTP;
                                    const ASourceName,ADestDir,ADestName : string;
                                    AEncryption : TEncryption; AOptions : TCopyOptions;
                                    ASuspend : Boolean; ABufSize : integer;
                                    APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FFtp:=AFtp;
  FCompress:=coCompress in AOptions; FZipCompress:=coZipCompress in AOptions;
  FVerify:=coVerify in AOptions;
  if FZipCompress then FErrFlag:=errZip
  else if FCompress then FErrFlag:=errGZip
  else if assigned(FCrypting) then FErrFlag:=errEncrypt
  else FErrFlag:=errCopy;
  with FFtp do begin
    ValidFileList:=false;
    OnWork:=ShowWork;
    end;
  FDestDir:=ADestDir; FFileSize:=0;
  end;

destructor TWriteFtpThread.Destroy;
begin
  FFtp.OnWork:=nil;
  inherited Destroy;
  end;

procedure TWriteFtpThread.ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if AWorkMode=wmWrite then DoProgress(acFtpWrite,AWorkCount,true)
  else if AWorkMode=wmRead then DoProgress(acFtpRead,AWorkCount,true)
  end;

procedure TWriteFtpThread.CancelThread;
begin
  FUserBreak:=true;
  while not Done do begin
    Sleep(10);
    Application.ProcessMessages;
    end;
  with FFtp do if Connected then begin
    try Abort; except GetInternalResponse end;
    Sleep(200);
    end;
  end;

procedure TWriteFtpThread.EndThread (Error : integer);
begin
//  if (Error and errAllCodes)=errUserBreak then FFtp.Abort;
  inherited EndThread(Error);
  end;

{ ------------------------------------------------------------------- }
function TWriteFtpThread.WriteFile : integer;
var
  ec            : integer;
  FEnc,tff      : boolean;
  FTime         : TFileTime;
  utime         : cardinal;
  sGZip,sEnc    : TStream;
  fp            : int64;
  tzn,ten       : string;
begin
  ec:=errOK;
  if not FileExistsUNC(SourceName) then ec:=errError+errFileOpen;
  if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
  FTime:=GetFileLastWriteTime(SourceName);
  if ec=errOK then begin
    try
      sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
    except
      ec:=errError+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    FEnc:=assigned(FCrypting);
    tff:=sSource.Size>UMaxMemSize;
    if FCompress then begin
      DoProgress(acCompress,-sSource.Size,false);   // Reset progress indicator
      // unix time for gzip
      utime:=round(SecsPerDay*(FileTimeToDateTime(FTime)-25569));
      if tff then begin
        tzn:=FFtp.TempDir+tmpGzipName;
        sGZip:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sGZip:=TMemoryStream.Create;
      ec:=GzWriteHeader(sGZip,ExtractFilename(SourceName),utime,FEnc);
      fp:=sGZip.Position;
     // compress file
      if ec=errOK then begin
        if FEnc then begin
          if tff then begin
            ten:=FFtp.TempDir+tmpEncName;
            sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
            end
          else sEnc:=TMemoryStream.Create;
          ec:=GzCompress(sSource,sEnc,false);
          if ec=errOK then begin
          // z.B. wenn "sSource" gelockt ist (siehe LockFile)
            if UTotal<>sSource.Size then ec:=errError+errFileRead
            else begin
              DoProgress(acEncrypt,-sEnc.Size,false);   // Reset progress indicator
              sEnc.Seek(0,soFromBeginning);
              if (ec=errOK) and not FCrypting.EncryptStream(sEnc,sGZip,false) then
                ec:=errError+errFileECrypt;
              end;
            end;
          sEnc.Free;
          if tff then DeleteFile(ten);
          FCrc:=0;
          end
        else begin
          ec:=GzCompress(sSource,sGZip,false);
          // z.B. wenn "sSource" gelockt ist (siehe LockFile)
          if (ec=errOK) and (UTotal<>sSource.Size) then ec:=errError+errFileRead;
          end;
        end;
      if ec=errOK then begin
        CTotal:=sGZip.Position-fp;
//        if UTotal>MaxCardinal then UTotal:=MaxCardinal;
        if FEnc then ec:=GzWriteTrailer (sGZip,0)
        else ec:=GzWriteTrailer (sGZip,FCrc);
        if ec=errOK then begin
          DoProgress (acFtpWrite,-sGZip.Size,false);  // Reset
          sGZip.Seek(0,soFromBeginning);
          with FFtp do if ReConnect(FDestDir) then ec:=Flush(sGZip,DestName,false)
          else ec:=errFtpConnect;
          FFileSize:=sGZip.Size;
          if FUserBreak then begin
            Sleep(1000);     // auf Abort warten
            ec:=errUserBreak;
            end;
          if FVerify and (ec=errOK) then ec:=VerifyFile(sGZip);
          end;
        end;
      sGZip.Free;
      if tff then DeleteFile(tzn);
      end
    else begin
      if FEnc then begin
        DoProgress(acEncrypt,-sSource.Size,false);   // Reset progress indicator
        if tff then begin
          ten:=FFtp.TempDir+tmpEncName;
          sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
          end
        else sEnc:=TMemoryStream.Create;
        if not FCrypting.EncryptStream(sSource,sEnc,false) then ec:=errFileECrypt;
        if ec=errOK then begin
          DoProgress (acFtpWrite,-sEnc.Size,false);  // Reset
          sEnc.Seek(0,soFromBeginning);
          with FFtp do if ReConnect(FDestDir) then ec:=Flush(sEnc,DestName,false)
          else ec:=errFtpConnect;
          FFileSize:=sEnc.Size;
          end;
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        if FVerify and (ec=errOK) then ec:=VerifyFile(sEnc);
        sEnc.Free;
        if tff then DeleteFile(ten);
        end
      else begin
        DoProgress(acFtpWrite,-sSource.Size,false);   // Reset progress indicator
        ec:=FFtp.Flush(sSource,DestName,false);
        FFileSize:=sSource.Size;
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        if FVerify and (ec=errOK) then ec:=VerifyFile(sSource);
        end;
      end;
    try sSource.Free; except end;
    end;
  Result:=ec;
  end;

function TWriteFtpThread.VerifyFile (fs : TStream) : integer;
var
  tff    : boolean;
  sRead  : TStream;
  tzf    : string;
begin
  DoProgress(acVerify,-fs.Size,false);   // Reset progress indicator
  tff:=fs.Size>CMaxMemSize;
  if tff then begin
    tzf:=FFtp.TempDir+tmpCompName;
    sRead:=TFileStream.Create(tzf,fmCreate); // write to disk
    end
  else sRead:=TMemoryStream.Create;
  DoProgress(acFtpRead,-fs.Size,false);   // Reset progress indicator
  Result:=FFtp.Fill(DestName,sRead);  // read  from FTP to stream
  if Result=errOK then Result:=CompareStreams(fs,sRead,false);
  end;

procedure TWriteFtpThread.Execute;
var
  ec       : integer;
  FTime    : TFileTime;
  tff      : boolean;
  tzn      : string;
begin
  if FZipCompress then begin
    ec:=errOK;
    if not FileExistsUNC(SourceName) then ec:=errError+errFileOpen;
    if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
    FTime:=GetFileLastWriteTime(SourceName);
    if ec=errOK then begin
      try
        sSource:=TFileStream.Create(FilenameToUNC(SourceName),fmOpenRead+fmShareDenyNone);
      except
        ec:=errError+errFileOpen;
        end;
      end;
    if ec=errOK then begin
      tff:=sSource.Size>UMaxMemSize;
      if tff then begin
        tzn:=FFtp.TempDir+tmpZipName;
        sDest:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sDest:=TMemoryStream.Create;
      ec:=CompressAsZip (FCompress);
      if ec=errOK then begin
        DoProgress(acFtpWrite,-sSource.Size,false);   // Reset progress indicator
        ec:=FFtp.Flush(sDest,DestName,false);
        FFileSize:=sSource.Size;
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        end;
      try sDest.Free; except end;
      if tff then DeleteFile(tzn);
      end;
    try sSource.Free; except end;
    EndThread(ec);
    end
  else EndThread (WriteFile);
  end;

{ TReadFtpThread ------------------------------------------------------------- }
  // Read file via FTP with opt. encryption
constructor TReadFtpThread.Create (AFtp : TExtFTP;
                                   const ASourceName,ADestName : string;
                                   AFileSize : int64;
                                   AEncryption : TEncryption;
                                   ACompress,AZipCompress,ASuspend : Boolean;
                                   ABufSize : integer;
                                   APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FFtp:=AFtp; FCompress:=ACompress; FZipCompress:=AZipCompress;
  FFileSize:=AFileSize;
  if FZipCompress then FErrFlag:=errZip
  else if FCompress then FErrFlag:=errGUnZip
  else FErrFlag:=errCopy;
  FFtp.OnWork:=ShowWork;
  end;

destructor TReadFtpThread.Destroy;
begin
  FFtp.OnWork:=nil;
  inherited Destroy;
  end;

procedure TReadFtpThread.ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if AWorkMode=wmRead then DoProgress(acFtpRead,AWorkCount,true);
  end;

procedure TReadFtpThread.CancelThread;
begin
  FUserBreak:=true;
  while not Done do begin
    Sleep(10);
    Application.ProcessMessages;
    end;
  with FFtp do if Connected then begin
    try Abort; except GetInternalResponse end;
    end;
  end;

procedure TReadFtpThread.EndThread (Error : integer);
begin
//  if (Error and errAllCodes)=errUserBreak then FFtp.Abort;
  inherited EndThread(Error);
  end;

{ ------------------------------------------------------------------- }
function TReadFtpThread.ReadFile : integer;
var
  ec            : integer;
  FEnc,tff      : boolean;
  utime         : cardinal;
  sEnc          : TStream;
  CSize,USize,
  fp            : int64;
  tzn,ten,s     : string;
begin
  ec:=errOK;
  FEnc:=assigned(FCrypting);
  if (length(DestName)=0) then ec:=errError+errFileCreate;
  if ec=errOK then begin
    try
      sDest:=TFileStream.Create(DestName,fmCreate);
    except
      ec:=errError+errFileCreate;
      end;
    end;
  if ec=errOK then begin
    if FCompress then begin
      tff:=FFileSize>CMaxMemSize;
      DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
      if tff then begin
        tzn:=FFtp.TempDir+tmpGzipName;
        sSource:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sSource:=TMemoryStream.Create;
      ec:=FFtp.Fill(SourceName,sSource);  // read
      if ec=errOK then begin
        sSource.Seek(0,soFromBeginning);
        if FZipCompress then ec:=UnZipStream(sSource,sDest)
        else begin
          // uncompress file
          ec:=GzReadHeader(sSource,s,uTime,FEnc,CSize,USize);
          if ec=errOK then begin
            fp:=sSource.Position;
            if FEnc then begin
              DoProgress(acDecrypt,-sSource.Size,false);  // Reset progress indicator
              if tff then begin
                ten:=FFtp.TempDir+tmpEncName;
                sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
                end
              else sEnc:=TMemoryStream.Create;
    //          with sSource do n:=Size-Position-8;
              if FCrypting.DecryptStream(sSource,sEnc,CSize,false) then begin
                DoProgress(acUnCompress,-USize,false);  // Reset progress indicator
                sEnc.Seek(0,soFromBeginning);
                ec:=GzUncompress(sEnc,sDest,sEnc.size,false);
                end
              else ec:=errError+errFileDCrypt;
              sEnc.Free;
              if tff then DeleteFile(ten);
              // Ignore trailer
              end
            else begin
              DoProgress(acUnCompress,-USize,false);      // Reset progress indicator
              ec:=GzUnCompress(sSource,sDest,false);
              if (ec=errOK) then begin
    //            CTotal:=sSource.Position-fp;
                ec:=GzReadTrailer (sSource,Usize);
                end;
              end;
            end;
          end;
        end;
      sSource.Free;
      if tff then DeleteFile(tzn);
      end
    else begin
      tff:=FFileSize>UMaxMemSize;
      if FEnc then begin
        DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
        if tff then begin
          ten:=FFtp.TempDir+tmpEncName;
          sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
          end
        else sEnc:=TMemoryStream.Create;
        ec:=FFtp.Fill(SourceName,sEnc);  // read
        if ec=errOK then begin
          DoProgress (acDecrypt,-sEnc.Size,false);  // Reset
          sEnc.Seek(0,soFromBeginning);
          if not FCrypting.DecryptStream(sEnc,sDest,sEnc.Size,false) then ec:=errFileDCrypt;
          end;
        if (ec=errOK) and FUserBreak then ec:=errUserBreak;
        sEnc.Free;
        if tff then DeleteFile(ten);
        end
      else begin
        DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
        ec:=FFtp.Fill(SourceName,sDest);
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        end;
      end;
    try sDest.Free; except end;
    end;
  Result:=ec;
  end;

procedure TReadFtpThread.Execute;
begin
  EndThread (ReadFile);
  end;

{ TCompareFtpThread ------------------------------------------------------------- }
  // Compare file from FTP with local file with opt. encryption
function TCompareFtpThread.CompareFile : integer;
var
  ec            : integer;
  FEnc,tff,tuf  : boolean;
  sUnp,
  sGZip,sEnc    : TStream;
  CSize,USize,
  fp            : int64;
  tzn,ten,tun   : string;
begin
  ec:=errOK;
  FEnc:=assigned(FCrypting);
  if not FileExists(DestName) then ec:=errError+errNotFound;
  if ec=errOK then begin
    try
      sDest:=TFileStream.Create(DestName,fmOpenRead);
    except
      ec:=errError+errFileCreate;
      end;
    end;
  if ec=errOK then begin
    if FCompress then begin
      tff:=FFileSize>CMaxMemSize;
      if tff then begin
        tzn:=FFtp.TempDir+tmpGzipName;
        sGZip:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sGZip:=TMemoryStream.Create;
      DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
      ec:=FFtp.Fill(SourceName,sGZip);  // read from FTP to stream
      if ec=errOK then begin
        sGZip.Seek(0,soFromBeginning);
        if FZipCompress then begin
          tuf:=sDest.Size>CMaxMemSize;
          if tuf then begin
            tun:=FFtp.TempDir+tmpUzipName;
            sUnp:=TFileStream.Create(tun,fmCreate); // write to disk
            end
          else sUnp:=TMemoryStream.Create;
          ec:=UnZipStream(sGZip,sUnp)
          end
        else begin
          ec:=GzReadSizes(sGZip,CSize,USize);
          if ec=errOK then begin
            tuf:=USize>CMaxMemSize;
            if tuf then begin
              tun:=FFtp.TempDir+tmpUzipName;
              sUnp:=TFileStream.Create(tun,fmCreate); // write to disk
              end
            else sUnp:=TMemoryStream.Create;
            fp:=sGZip.Position;
            if FEnc then begin
              DoProgress(acDecrypt,-sGZip.Size,false); // Reset progress indicator
              if tuf then begin
                ten:=FFtp.TempDir+tmpEncName;
                sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
                end
              else sEnc:=TMemoryStream.Create;
    //          with sGZip do n:=Size-Position-8;
              if FCrypting.DecryptStream(sGZip,sEnc,CSize,false) then begin
                DoProgress(acUnCompress,-USize,false); // Reset progress indicator
                sEnc.Seek(0,soFromBeginning);
                ec:=GzUncompress(sEnc,sUnp,sEnc.size,false);
                end
              else ec:=errError+errFileDCrypt;
              sEnc.Free;
              if tuf then DeleteFile(ten);
              // Ignore trailer
              end
            else begin
              DoProgress(acUnCompress,-USize,false);  // Reset progress indicator
              ec:=GzUnCompress(sGZip,sUnp,false);
              if (ec=errOK) then ec:=GzReadTrailer (sGZip,USize);
              end;
            end;
          end;
        if ec=errOK then begin  // compare
          ec:=CompareStreams(sDest,sUnp,false);
          end;
        sUnp.Free;
        if tuf then DeleteFile(tun);
        end;
      sGZip.Free;
      if tff then DeleteFile(tzn);
      end
    else begin
      tuf:=FFileSize>UMaxMemSize;
      if tuf then begin
        tun:=FFtp.TempDir+tmpUzipName;
        sUnp:=TFileStream.Create(tun,fmCreate); // write to disk
        end
      else sUnp:=TMemoryStream.Create;
      if FEnc then begin
        DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
        if tuf then begin
          ten:=FFtp.TempDir+tmpEncName;
          sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
          end
        else sEnc:=TMemoryStream.Create;
        ec:=FFtp.Fill(SourceName,sEnc);  // read
        if ec=errOK then begin
          DoProgress (acDecrypt,-sEnc.Size,false);  // Reset
          sEnc.Seek(0,soFromBeginning);
          if not FCrypting.DecryptStream(sEnc,sUnp,sEnc.Size,false) then ec:=errFileDCrypt;
          end;
        if (ec=errOK) and FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        sEnc.Free;
        if tuf then DeleteFile(ten);
        end
      else begin
        DoProgress(acFtpRead,-FFileSize,false);   // Reset progress indicator
        ec:=FFtp.Fill(SourceName,sUnp);
        if (ec=errOK) and FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        end;
      if ec=errOK then ec:=CompareStreams(sDest,sUnp,false);
      sUnp.Free;
      if tuf then DeleteFile(tun);
      end;
    try sDest.Free; except end;
    end;
  Result:=ec;
  end;

procedure TCompareFtpThread.Execute;
begin
  EndThread (CompareFile);
  end;

end.
