(* Delphi-Unit
   Unterroutinen und Komponenten fr Unicode
   =========================================

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Version 1.0 - Feb 2009
           1.1 - May 2010
   *)

unit UnicodeUtils;

interface

uses System.Classes, System.SysUtils, System.IniFiles;

const
  UcBom = $FEFF;   //UniCode-Signatur (Low Endian)
  Utf8BomS : RawByteString = #$EF#$BB#$BF;
  Utf8Bom : array[0..2] of byte = ($EF,$BB,$BF);

type
  TUnicodeIniFile = class(TIniFile)
    FFileName : string;
    constructor CreateForRead (const FileName: string);
    constructor CreateForWrite (const FileName: string);
    function ReadString(const Section, Ident, Default: string): string; override;
  end;

function IsAnsiStr (ws : string) : boolean;

// Functions to extend full paths to UNC path (==> pathlength >260)
function FilenameToUNC(const FileName : string) : string;

function ForceDirectoriesUNC(const DirName : string) : boolean;
function RemoveDirUNC(const DirName : string) : integer;
//function SetCurrentDirUNC(const DirName : string) : boolean;
function DirectoryExistsUNC(Directory : string; FollowLink: Boolean = True) : boolean;

function FileExistsUNC(const FileName : string; FollowLink: Boolean = True) : boolean;
function DeleteFileUNC(const FileName : string; DelReadOnly : boolean = false) : integer;
function EraseFileUNC(const FileName : string; DelReadOnly : boolean = false) : boolean;
function RenameFileUNC(const OldName,NewName : string) : integer; overload;
function RenameFileUNC(const OldName,NewName : string; Attr : integer) : integer; overload;
function ExpandFileNameUNC(const FileName: string): string;

function LongFileSizeUNC (const FileName : string) : int64;
function FileGetAttrUNC(const FileName: string; FollowLink: Boolean = True): Integer;
function FileSetAttrUNC(const FileName: string; Attr: Integer; FollowLink: Boolean = True): Integer;
function FileChangeAttrUNC (const FileName: string; Attr: Integer; Clear : boolean;
                              FollowLink: Boolean = True): Integer;
function FileAgeUNC(const FileName: string; Default : double) : TDateTime;
function DirectoryAgeUNC(const FileName: string; Default : double) : TDateTime;

function WideUpCase (c : WideChar) : WideChar;

function WriteUtf8Bom (const Filename : string) : integer;

implementation

uses Winapi.Windows, FileUtils, WinUtils;

constructor TUnicodeIniFile.CreateForRead (const FileName: string);
begin
  FFileName:=FileName;
  inherited Create(FileName);
  end;

constructor TUnicodeIniFile.CreateForWrite (const FileName: string);
const
  UcSection = '[Unicode]';
var
  fs  : TFileStream;
  uid : word;
  sl  : TStringList;
begin
  FFileName:=FileName;
  try
    if FileExists(FileName) then begin
      fs:=TFileStream.Create(FileName,fmOpenReadWrite);
      fs.Read(uid,2);
      if uid<>UcBom then begin  // in Unicode konvertieren
        fs.Position:=0;
        sl:=TStringList.Create;
        sl.LoadFromStream(fs);
        sl.Insert(0,UcSection);
        fs.Size:=0;
        sl.SaveToStream(fs,TEncoding.Unicode);
        sl.Free;
        end
      end
    else begin
      fs:=TFileStream.Create(FileName,fmCreate);
      sl:=TStringList.Create;
      sl.Add(UcSection);
      sl.SaveToStream(fs,TEncoding.Unicode);
      sl.Free;
      end;
  finally
    fs.Free;
    end;
  inherited Create(FileName);
  end;

// replace original function to increase the string length to > 2047
function TUnicodeIniFile.ReadString(const Section, Ident, Default: string): string;
var
  Buffer: array[0..65535] of Char;
begin
  SetString(Result, Buffer, GetPrivateProfileString(PChar(Section),
    PChar(Ident), PChar(Default), Buffer, Length(Buffer), PChar(FFileName)));
  end;

{ ------------------------------------------------------------------- }
function IsAnsiStr (ws : string) : boolean;
var
  i : integer;
begin
  Result:=false;
  for i:=0 to length(ws) do if TWord(ws[i]).Hi<>0 then exit;
  Result:=true;
  end;

{ ------------------------------------------------------------------- }
function FilenameToUNC(const FileName : string) : string;
begin
  if (copy(FileName,1,2)='\\') then begin
    if (copy(FileName,3,2)='?\') then Result:=FileName               // is already UNC
//    else  Result:=Filename;
    else Result:='\\?\UNC\'+Copy(Filename,3,length(Filename)-2);     // network
    end
  else if ((length(FileName)>0) and (FileName[2]<>':')) then
    Result:=Filename                                                 // relative path
  else Result:='\\?\'+FileName;                                      // local
  end;

{ ------------------------------------------------------------------- }
function ForceDirectoriesUNC(const DirName : string) : boolean;
begin
  Result:=ForceDirectories(FilenameToUNC(DirName));
  end;

function RemoveDirUNC(const DirName : string) : integer;
begin
  if RemoveDirectory(PChar(FilenameToUNC(DirName))) then Result:=NO_ERROR
  else Result:=GetLastError;
  end;

{function SetCurrentDirUNC(const DirName : string) : boolean;
begin
  Result:=SetCurrentDirectory(PChar(FilenameToUNC(DirName)));  // UNC paths not supported
end;   }

function DirectoryExistsUNC (Directory : string; FollowLink: Boolean = True) : boolean;
// replaces "DirectoryExists" in SysUtils
// bug fix: acknowledge ERROR_BAD_NET_NAME and ERROR_NOT_READY as possible error
var
  Code: Cardinal;
  Handle: THandle;
  LastError: Cardinal;
begin
  Result := False;
  Directory:=FilenameToUNC(Directory);
  Code:=GetFileAttributes(PChar(Directory));
  if Code <> INVALID_FILE_ATTRIBUTES then begin
    if faSymLink and Code = 0 then Result := faDirectory and Code <> 0
    else begin
      if FollowLink then begin
        Handle := CreateFile(PChar(Directory), GENERIC_READ, FILE_SHARE_READ, nil,
          OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0);
        if Handle <> INVALID_HANDLE_VALUE then begin
          CloseHandle(Handle);
          Result := faDirectory and Code <> 0;
          end;
        end
      else if faDirectory and Code <> 0 then Result := True
      else begin
        Handle := CreateFile(PChar(Directory), GENERIC_READ, FILE_SHARE_READ, nil,
          OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0);
        if Handle <> INVALID_HANDLE_VALUE then begin
          CloseHandle(Handle);
          Result := False;
          end
        else Result := True;
        end;
      end;
    end
  else begin
    LastError := GetLastError;
    Result := (LastError <> ERROR_FILE_NOT_FOUND) and
      (LastError <> ERROR_PATH_NOT_FOUND) and
      (LastError <> ERROR_INVALID_NAME) and
      (LastError <> ERROR_BAD_NET_NAME) and
      (LastError <> ERROR_NOT_READY) and
      (LastError <> ERROR_BAD_NETPATH);
    end;
  end;

{ ------------------------------------------------------------------- }
function FileExistsUNC(const FileName : string; FollowLink: Boolean = True) : boolean;
begin
  Result:=FileExists(FilenameToUNC(FileName),FollowLink);
  end;

function DeleteFileUNC(const FileName : string; DelReadOnly : boolean = false) : integer;
begin
  if DelReadOnly then Result:=FileChangeAttrUNC(Filename,faReadOnly,true,false)
  else Result:=NO_ERROR;;
  if Result=NO_ERROR then begin
    if DeleteFile(PChar(FilenameToUNC(FileName))) then Result:=NO_ERROR
    else Result:=GetLastError;
    end;
  end;

function EraseFileUNC(const FileName : string; DelReadOnly : boolean = false) : boolean;
begin
  Result:=DeleteFileUNC(FileName,DelReadOnly)=NO_ERROR;
  end;

function RenameFileUNC(const OldName,NewName : string) : integer;
begin
  if RenameFile(FilenameToUNC(OldName),FilenameToUNC(NewName)) then Result:=NO_ERROR
  else Result:=GetLastError;
  end;

// Result = 0 : ok
//        > 0 : sytem error code - rename failed
//        < 0 : sytem error code - set attribute failed
function RenameFileUNC(const OldName,NewName : string; Attr : integer) : integer;
begin
  Result:=RenameFileUNC(OldName,NewName);
  if Result=NO_ERROR then Result:=-FileSetAttrUNC(NewName,Attr,true);
  end;

function ExpandFileNameUNC(const FileName: string): string;
var
  FName: PChar;
  Buffer: array[0..32767] of Char;
  Len: Integer;
begin
  Len:=GetFullPathName(PChar(FilenameToUNC(FileName)), Length(Buffer), Buffer, FName);
  if Len <= Length(Buffer) then SetString(Result, Buffer, Len)
  else if Len > 0 then begin
    SetLength(Result,Len);
    Len := GetFullPathName(PChar(FileName), Len, PChar(Result),FName);
    if Len<Length(Result) then SetLength(Result,Len);
    end;
  end;

{ ------------------------------------------------------------------- }
function LongFileSizeUNC (const FileName : string) : int64;
begin
  Result:=LongFileSize(FilenameToUNC(FileName));
  end;

function FileGetAttrUNC(const FileName: string; FollowLink: Boolean = True): Integer;
begin
  Result:=FileGetAttr(FilenameToUNC(FileName),FollowLink);
  end;

function FileSetAttrUNC(const FileName: string; Attr: Integer; FollowLink: Boolean = True): Integer;
begin
  Result:=FileSetAttr(FilenameToUNC(FileName),Attr,FollowLink);
  end;

// Clear = true  : clear given attributes
//       = false : set given attributes
function FileChangeAttrUNC (const FileName: string; Attr: Integer; Clear : boolean;
                            FollowLink: Boolean = True) : Integer;
var
  ao,an : integer;
begin
  ao:=FileGetAttr(FilenameToUNC(FileName),FollowLink);
  if Clear then an:=ao and not Attr else an:=ao or Attr;
  if ao<>an then Result:=FileSetAttr(FilenameToUNC(FileName),an,FollowLink)
  else Result:=NO_ERROR;
  end;

function FileAgeUNC(const FileName: string; Default : double) : TDateTime;
begin
  if not FileAge(FilenameToUNC(FileName),Result,false) then Result:=Default;
  end;

function DirectoryAgeUNC(const FileName: string; Default : double) : TDateTime;
begin
  if not DirectoryAge(FilenameToUNC(FileName),Result,false) then Result:=Default;
  end;

{ ------------------------------------------------------------------- }
function WideUpCase (c : WideChar) : WideChar;
var
  s : string;
begin
  s:=c;
  CharUpperBuff(PChar(s),1);
  Result:=s[1];
  end;

{ ------------------------------------------------------------------- }
function WriteUtf8Bom (const Filename : string) : integer;
var
  fp    : file;
  n     : integer;
begin
  assignfile(fp,Filename);
  {$I-} rewrite(fp,1) {$I+};
  Result:=IOResult;
  if Result=0 then begin
    BlockWrite(fp,Utf8Bom,3,n);
    CloseFile(fp);
    end;
  end;

end.
